/*
 * WSHierarchy.cpp - Watershed hierarchy
 *
 * Computes the watershed hierarchy of a 4 adjacency gradient.
 * Given an image of size h*w = N
 * The inputs consists of:
 *      - an uint8 array of size 1*N that contains the edge values (see demo file) 
 *      - the heigth h
 *      - the width w
 *      - the type of hierarchy to compute (0=Area, 1=Dynamics, 2=Volume) 
 * The outputs are
 *      - an utin8 array of size 1*N that contains the saliency map of the hierarchy (see demo file) 
 *      - the execution time of the inner function in ms (without boilerplate code)
 *
 * The calling syntax is:
 *
 *		[sm, time] = WSHierarchy(edges, height, width, type)
 *
 * This is a MEX file for MATLAB.
 *
 * Author: Benjamin Perret ESIEE Paris 2017
 *
*/


#include <matrix.h>
#include <mex.h>

#include "Wrapper/mx_image.hpp"
#include "src/mccodimage.h"
#include "src/jccodimage.h"
#include "src/mcimage.h"
#include "src/jcimage.h"
#include "src/lhierarchie.h"
#include <time.h>
#include <iostream>

extern "C" bool mxUnshareArray(mxArray *array_ptr, bool noDeepCopy);




void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[])
{
    // check the params
    if (nrhs != 4)
        mexErrMsgTxt("This function takes 4 inputs");
    if (nlhs <1)
        mexErrMsgTxt("This function gives at least 2 output");
    if (nlhs > 2)
        mexErrMsgTxt("This function gives at most 2 output");
       

    
    // inplace editing
    //if(!mxUnshareArray(const_cast<mxArray *>(prhs[0]), true))
    //    mexErrMsgTxt("Inplace failed");
    
    // read the inputs
    MxImage<uint8_t> x( prhs[0] );
    const int h = (int) mxGetScalar( prhs[1] );
    const int w = (int) mxGetScalar( prhs[2] );
    const int type = (int) mxGetScalar( prhs[3] );
    
    
    
    
    int N = x.total();
    if(N != 2*h*w)
         mexErrMsgTxt("Incorrect size for edges");
    
    MxImage<uint8_t> result( plhs[0], 1,N);
    x.copy_to(result);
    /* Strucutre wrapping */
    struct xvimage  g;
    
    
    g.image_data = result.begin();
    g.name = NULL;
    rowsize(&g) = w;
    colsize(&g) = h;
    depth(&g) = 1;
    datatype(&g) = 2;
    g.xdim = g.ydim = g.zdim = 0.0;
    
    /* Type identification */
    int param;
    switch(type)
    {
        case 0: param = SURFACE; break;
        case 1: param = DYNAMIC; break;
        case 2: param = VOLUME; break; 
        default:  mexErrMsgTxt("Incorrect hierarchy type.");
    }
    
    struct timespec start, end;
    clock_gettime(CLOCK_MONOTONIC_RAW, &start);

    if (saliencyGa(&g, param, nullptr) != 1)
    {

        mexErrMsgTxt("WS failed.");
        
    }
    
    clock_gettime(CLOCK_MONOTONIC_RAW, &end);
    
    if(nlhs>1)
    {
        uint64_t delta_ms = ((end.tv_sec - start.tv_sec) * 1000000 + (end.tv_nsec - start.tv_nsec) / 1000)/1000;
        plhs[1] = mxCreateDoubleScalar((double)delta_ms);
       
    }
  
}