#!/usr/bin/env python3

# Copyright ESIEE (2016)
#
# benjamin.perret@esiee.fr
#
# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.



from multiprocessing import Pool, Lock
import os
from time import sleep
import glob
import shutil
import sys
import stat
import json
import logging
import subprocess
import tqdm
import shlex
singleTaskTimeout = 2200 # seconds! 

baseDirectory = os.getcwd() + "/"
databaseDirectory = baseDirectory + "database/"
libDirectory = baseDirectory + "lib/"
logDirectory = baseDirectory + "log/"

kevalScoringDirectory = baseDirectory + "WatershedsGenericAndTools/linux/bin/"

errorLogFile = logDirectory + "errors.log"

logger = None
loggerLock = Lock()
def initLogger(file, cleanup):
    global logger
    if cleanup:
        mode='w'
    else:
        mode='a'
    logger = logging.getLogger("HEvalLogger")
    fileHandler = logging.FileHandler(file,mode=mode)
    
    handler = logging.StreamHandler()
    formatter = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")
    
    handler.setFormatter(formatter)
    fileHandler.setFormatter(formatter)
    logger.removeHandler(handler)
    #logger.addHandler(handler)
    logger.addHandler(fileHandler)
    

def logError(msg):
    loggerLock.acquire()
    logger.error(msg)
    loggerLock.release()


def isEmpty(file):
    return os.stat(file).st_size==0

def prepareResultDir(directory, forceRecompute):
    if os.path.exists(directory) and forceRecompute:
        shutil.rmtree(directory)
    if not os.path.exists(directory):
        os.makedirs(directory)

def adjustPool(poolr, nbThreads):
    pool = poolr[0]
    if pool._processes != nbThreads:
        pool.close()
        pool.join()
        pool = Pool(processes = nbThreads)
        poolr[0] =  pool
    

def runCommand(cwdCom):

    directory = cwdCom[0]
    command = cwdCom[1]
    if not directory:
        directory = "."
    if directory[0] == ".":
        directory = baseDirectory + directory
    err=1
    args= [directory + "/" + command[0]]
    
    os.chmod(args[0], stat.S_IRWXU | stat.S_IRGRP | stat.S_IXGRP | stat.S_IROTH | stat.S_IXOTH)
    for i in range(1, len(command)):
        args.extend(shlex.split(command[i]))
    #print(args)
    try:
        proc = subprocess.run(args, stdout=subprocess.PIPE, stderr=subprocess.PIPE, cwd=directory, timeout=singleTaskTimeout, check=True, universal_newlines=True)
        #subprocess.run(args,  cwd=directory, timeout=singleTaskTimeout, check=True)
        err=0
    except subprocess.TimeoutExpired as e:
        logError("Timeout on " + directory + " " + str(command) + "\n------------ STDOUT:\n" + e.stdout + "\n------------ STDERR:\n" + e.stderr)
    except subprocess.CalledProcessError as e:
        logError("Exit failure " + directory + " " + str(command) + "\n------------ STDOUT:\n" + e.stdout + "\n------------ STDERR:\n" + e.stderr)
    except Exception as e:
        logError("Exception on " + directory + " " + str(command) + "\tMessage:" +str(e))
    return err

def validateConfiguration(conf=None, minimal=False):
    if not conf:
        conf = {}
    currentDirectory = os.getcwd()
    jobList = ["generateGraphFiles", "FOC", "FHC", "objectDetection"]
    
    # global configuration initia
    globalConf = conf.setdefault("global", {})
    globalConf.setdefault("forceRecompute", False)
    globalConf.setdefault("threads", 8)
    globalConf.setdefault("tmpFolder", baseDirectory + "tmp/")
    globalConf.setdefault("resultsFolder", baseDirectory + "results/")
    globalConf.setdefault("databases", ["grabcut", "weizmann1obj", "bsds500"])
    globalConf.setdefault("imageSet", "test")
    globalConf.setdefault("logFile", baseDirectory + "log/errors.log")
    globalConf.setdefault("clearLog", False)
    globalConf.setdefault("cleanup", False)
    globalConf.setdefault("scripts", [])
    
    for job in jobList:
        jobconf = conf.setdefault(job, {})
        jobconf.setdefault("enabled", not not jobconf)
        
    conf["FOC"].setdefault("measures", ["BCE"])
    conf["FHC"].setdefault("measures", ["BCE"])
    conf["FOC"].setdefault("databases", ["bsds500"])
    conf["FHC"].setdefault("databases", ["bsds500"])

    conf["objectDetection"].setdefault("databases", ["grabcut", "weizmann1obj"])
    
    if not minimal:
        for job in jobList:
            for key in ["forceRecompute", "threads", "databases", "tmpFolder", "resultsFolder",  "imageSet", "scripts"]:
                conf[job].setdefault(key, globalConf[key])
    
    return conf

  
    
            
            
def runasync(poolr,args,desc, timeout=2200):

    global singleTaskTimeout
    pool = poolr[0]
    singleTaskTimeout=timeout #yeurk
    total = len(args)
    err=0
    if total>0:
        #pool.starmap(runCommand, all, 20)
        nbt = pool._processes
        chunkSize = 1#max(1,min(20, total//nbt))
        #print("Multiprocessing chunk size: " + str(chunkSize))
        res = pool.imap(runCommand, args, chunkSize)
        bar = tqdm.tqdm(total=total, desc=desc, leave=True)
        
        for i in range(total):
            err+=next(res)
            bar.update(1)
        bar.close()
    else: #nice !
        bar = tqdm.tqdm(total=1, desc=desc, leave=True)
        bar.update(1)
        bar.close()
    return err



# ####################################################################################
#  Graph file creation


def launchGraphFileCreationParralel(pool, conf):
    if not conf["enabled"]:
        return 0
    adjustPool(pool, conf["threads"])
    all = []
    for database in conf["databases"]:
        dbImageDir = databaseDirectory + database + "/data/images/"
        idsFile=databaseDirectory + database + "/ids_" + conf["imageSet"] + ".txt"
        with open(idsFile, "r") as f:
            ids = f.readlines()
        for script in conf["scripts"]:
            scriptDirectory, _, scriptName = script.rpartition("/")
            outDir = conf["tmpFolder"]  + scriptName +"/" + database + "/GRAPH/"
            prepareResultDir(outDir, conf["forceRecompute"])
            for image in ids:
                image = image.strip()
                outFile = outDir + image + ".graph"
                if not os.path.isfile(outFile):                
                    inFile = dbImageDir + image +".ppm"
                    all.append( (scriptDirectory, [scriptName, inFile, outFile]) )

    return runasync(pool,all, desc="Graph files creation")






# ####################################################################################
#  Kcurve evaluation

def launchScoring(pool, conf, threshold=False):
    if not conf["enabled"]:
        return 0
    adjustPool(pool, conf["threads"])
    all = []
    maxkKArg =' -maxK 200 '
    if threshold:
        actionArg = " -action threshold "
        suffixpath = "_threshold/"
    else:
        actionArg = " -action upperBound "
        suffixpath = "/"
    for database in conf["databases"]:
        dbImageDir = databaseDirectory + database + "/data/images/"
        gtImageDir = databaseDirectory + database + "/data/groundTruth/"

        
        for script in conf["scripts"]:
                scriptDirectory, _, scriptName = script.rpartition("/")
                for measure in conf["measures"]:
                    outDir = conf["tmpFolder"] + scriptName + "/" + database + "/"  + measure + suffixpath
                    prepareResultDir(outDir, conf["forceRecompute"])
        
        idsFile = databaseDirectory + database + "/ids_" + conf["imageSet"] + ".txt"
        with open(idsFile, "r") as f:
            ids = f.readlines()
        

        for image in ids:
            image = image.strip()
            gtFiles = glob.glob(gtImageDir + image + "_*.pgm")
            gtArgs = " -gt " + " -gt ".join(gtFiles)

            for script in conf["scripts"]:
                scriptDirectory, _, scriptName = script.rpartition("/")
                tmpDir = conf["tmpFolder"] + scriptName + "/" + database + "/GRAPH/" 
                inFileArg = " -i " + tmpDir + image +".graph"
                for measure in conf["measures"]:
                    measureArg = " -score " + measure
                    outDir = conf["tmpFolder"] + scriptName + "/" + database + "/"  + measure + suffixpath
                    outFile =  outDir + image + ".json"
                    if not os.path.isfile(outFile):
                        outFileArg =  " -o " + outDir + image + ".json"
                        args = " ".join([maxkKArg,inFileArg,outFileArg,gtArgs,actionArg,measureArg])
                        all.append( (kevalScoringDirectory, ["evalcombiPartition", args]) )

    err = runasync(pool,all, desc="KEval scoring (threshold= " +str(threshold) +")")
    err += launchCollectMeasures(pool,conf, threshold)
    return err

def launchCollectMeasures(pool,conf, threshold=False):
    if not conf["enabled"]:
        return 0
    if threshold:
        suffixpath = "_threshold"
    else:
        suffixpath = ""
    all = []

    for database in conf["databases"]:
        idsFile=databaseDirectory + database + "/ids_" + conf["imageSet"] + ".txt"
        for script in conf["scripts"]:
            scriptDirectory, _, scriptName = script.rpartition("/")
            for measure in conf["measures"]:
                outDir = conf["tmpFolder"] + scriptName + "/" + database + "/"  + measure + suffixpath +"/"
                resultDir = conf["resultsFolder"] + scriptName + "/"
                prepareResultDir(resultDir , False)
                resultFile = resultDir + scriptName + "_"  + database + "_"  + measure + suffixpath + ".txt"
                if not os.path.isfile(resultFile) or conf["forceRecompute"]:
                    all.append((libDirectory,["CollectEvalKCurvePartition.py", outDir,idsFile, resultFile]))
    return runasync(pool,all, desc="KEval score collection (threshold= " +str(threshold) +")")



# ####################################################################################
#  Marker evaluation

def scoreMarkerFile(inFile, markerFile, gtFile, end):
    command = baseDirectory + "keval/linux/bin/evalMarkedSegmentation " +inFile + " " + markerFile + " "  +gtFile + " " +end
    os.system(command)

def launchScoringMarker(pool,conf):
    if not conf["enabled"]:
        return 0
    adjustPool(pool, conf["threads"])
    
    all = []
    for database in conf["databases"]:
        #dbImageDir = databaseDirectory + database + "/data/images/"
        markerDir = databaseDirectory + database + "/data/markers/"
        gtImageDir = databaseDirectory + database + "/data/groundTruth/"
        idsFile = databaseDirectory + database + "/ids_" + conf["imageSet"] + ".txt"
        with open(idsFile, "r") as f:
            ids = f.readlines()
        
        for script in conf["scripts"]:
            scriptDirectory, _, scriptName = script.rpartition("/")
            #tmpDir = conf["tmpFolder"]  + scriptName +"/" + database +"/MAT/"

            for marker in [ "markerll", "markerfs", "markerss"]:
                outDir = conf["tmpFolder"]  + scriptName +"/" + database +"/marker/" + marker + "/"
                prepareResultDir(outDir, conf["forceRecompute"])

        for image in ids:
            image = image.strip()

            gtFile = gtImageDir + image + ".pgm"
                
            for script in conf["scripts"]:
                scriptDirectory, _, scriptName = script.rpartition("/")
                
                tmpDir = conf["tmpFolder"]  + scriptName +"/" + database + "/GRAPH/"
                inFile = tmpDir + image +".graph"
                for marker in [ "markerll", "markerfs", "markerss"]:
                    outDir = conf["tmpFolder"]  + scriptName +"/" + database +"/marker/" + marker + "/"
                    markerFile = markerDir + image + "_" + marker + ".pgm"
                    outFile =  outDir + image  +".txt"
                    if not os.path.isfile(outFile):
                        #all.append( (inFile, markerFile, gtFile, " > " + outFile) )
                        all.append( (kevalScoringDirectory, ["evalMarkedSegmentation", inFile, markerFile, gtFile, outFile]) )
    err = runasync(pool,all, desc="Marker scoring")
    err += launchCollectMarkerMeasures(pool,conf)
    
    return err


def launchCollectMarkerMeasures(pool,conf):

    all=[]
    for database in conf["databases"]:
        idsFile = databaseDirectory + database + "/ids_" + conf["imageSet"] + ".txt"
        for script in conf["scripts"]:
            scriptDirectory, _, scriptName = script.rpartition("/")
            for marker in [ "markerll", "markerfs", "markerss"]:
                inDir = conf["tmpFolder"]  + scriptName +"/" + database +"/marker/" + marker + "/"
                outFile = conf["resultsFolder"] + scriptName + "/" + scriptName + "_" + database + "_MARKER_" + marker  + ".txt" 
                if not os.path.isfile(outFile) or conf["forceRecompute"]:
                    all.append((libDirectory, ("CollectEvalMarker.py", inDir, idsFile, outFile) ))
    return runasync(pool,all, desc="MARKER scores collection")
                

# ####################################################################################
#  CleanUp

def doCleanUp(conf):
    for script in conf["scripts"]:
        scriptDirectory, _, scriptName = script.rpartition("/")
        tmpDir = conf["tmpFolder"]  + scriptName +"/"
        shutil.rmtree(tmpDir)
# ####################################################################################
#  Main

def main():
    
    confFile = "conf.json"
    if len(sys.argv) == 2:
        if sys.argv[1] == "default":
            conf = validateConfiguration(minimal=True)
            print(json.dumps(conf, indent=4))
            exit(0)
        confFile = sys.argv[1]
    try:
        with open(confFile) as data_file:    
            config = json.load(data_file)
    except Exception as e:
        print("ERROR reading configuration file: " + str(e))
        print("Use parameter 'default' to dump a sample configuration file to standard output")
        exit()
        
    
    
    validateConfiguration(config)
    initLogger(config["global"]["logFile"], config["global"]["clearLog"])
    errors=0
    
    pool = [Pool(processes=config["global"]["threads"])] #list wrapping to simulate pass by ref...
    
    errors += launchGraphFileCreationParralel(pool,config["generateGraphFiles"])
    errors += launchScoring(pool,config["FOC"],False)
    errors += launchScoring(pool,config["FHC"],True)
    errors += launchScoringMarker(pool,config["objectDetection"])

    
    if errors != 0:
        print("WARNING: " + str(errors) + " error detected!\n\tError log in log/errors.log" , file=sys.stderr)
        if config["global"]["cleanup"]:
            print("INFO: cleanup option aborded")
    else:
        if config["global"]["cleanup"]:
            doCleanUp(config["global"])
    pool[0].close()
    pool[0].join()
    


if __name__ == '__main__':
    main()





