#!/usr/bin/env python3

# Copyright ESIEE (2016)
#
# benjamin.perret@esiee.fr
#
# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

import sys
import json
import os
import csv
import numpy
import scipy
import itertools
import matplotlib.pyplot as plt
import scipy.interpolate
import scipy.signal
import matplotlib.cm as cmx
import matplotlib.colors as colors

const_kmin=1
const_kmax=100
const_kres=100
const_knmin=0.06 # 0 is not a valid fragmentation !
const_knmax=2
const_knres=100

def get_cmap(N):
    '''Returns a function that maps each index in 0, 1, ... N-1 to a distinct
    RGB color.'''
    color_norm  = colors.Normalize(vmin=0, vmax=N-1)
    scalar_map = cmx.ScalarMappable(norm=color_norm, cmap='hsv')
    def map_index_to_rgb_color(index):
        return scalar_map.to_rgba(index)
    return map_index_to_rgb_color



def computeScores(f, xmin, xmax, resolution):

    #op = scipy.integrate.quad(f, xmin, xmax,full_output=1)

    #v = op[0]
    #areaUnderCurve = v/(xmax-xmin)

    x = numpy.linspace(xmin, xmax, num=resolution, endpoint=True)
    y = f(x)

    y[y < 0] = 0 #avoid extrapolation fail

    v = numpy.trapz(y, x=x)
    areaUnderCurve = v/(xmax-xmin)

    imax = 0
    vmax = y[0]

    for i in range(len(x)):
        if vmax==0 or (y[i]-vmax)/vmax > 0.01:
            imax = i
            vmax = y[i]

    #imax = numpy.argmax(y)
    bestScoreY = y[imax]
    bestScoreX = x[imax]

    return areaUnderCurve, bestScoreX, bestScoreY




def extract_best_match_GT(data):
    nbgt = data["NUM_GT"]
    res = []
    for i in range(nbgt):
        datagt = data["RESULTS"][i]
        nbrgt = datagt["NUM_REGIONS"]
        values = datagt["VALUES"]
        file = datagt["FILE"].split("/")[-1]
        maxk = values[0][0]
        maxscore = values[0][1]
        for j in range(len(values)):
            if (values[j][1]-maxscore)/maxscore > 0.01:
                maxscore = values[j][1]
                maxk = values[j][0]

        res.append((nbrgt, maxk, maxscore, file))

        # xx, yy = zip(*values)
        # plt.plot(xx,yy)
        # plt.scatter((maxk),(maxscore))
        # plt.show()

    return res

def extract_score_atGTRegions(data):
    nbgt = data["NUM_GT"]
    res = []
    for i in range(nbgt):
        datagt = data["RESULTS"][i]
        nbrgt = datagt["NUM_REGIONS"]
        values = datagt["VALUES"]
        file = datagt["FILE"].split("/")[-1]
        maxk = values[0][0]
        maxscore = values[0][1]
        for j in range(len(values)):
            if abs(maxk-nbrgt) > abs(values[j][0]-nbrgt):
                maxscore = values[j][1]
                maxk = values[j][0]

        res.append((nbrgt, maxk, maxscore,file))

    return res


def filter(x,y):

    a = numpy.asarray((x,y))
    a = scipy.signal.savgol_filter(a, 51, 3)



    return  a[0], a[1] #


def filter2(funs, startPoint, endPoint, resolution):
    s = resolution
    nf = len(funs)
    x = numpy.linspace(startPoint, endPoint, num=s, endpoint=True)
    data = numpy.zeros((nf,s))
    for i in range(nf):
        data[i] = funs[i](x)

    y = numpy.mean(data, 0)

    return x, y

def augmentCurveLeftMin(x,y, maxV):
    nx = []
    ny = []
    curPos = 0
    curVal = y[0]
    nbe = len(x)
    for i in range(1,maxV+1):
        nx.append(i)
        if curPos < nbe and x[curPos] == i:
            curVal = y[curPos]
            curPos += 1
        ny.append(curVal)

    return nx, ny


def compute_mean_curve(data):
    global const_kmin, const_kmax, const_kres, const_knmin, const_knmax, const_knres
    nbgt = data["NUM_GT"]
    cmap = get_cmap(nbgt*2)
    funs = []
    normalizedFuns = []
    for i in range(nbgt):
        datagt = data["RESULTS"][i]
        nbrgt = datagt["NUM_REGIONS"]
        values = datagt["VALUES"]
        linearx = []
        lineary = []

        for j in range(len(values)):
            linearx.append(values[j][0])
            lineary.append(values[j][1])


        linearx, lineary = augmentCurveLeftMin(linearx, lineary, const_kmax)

        normalizedx = [] #[0]
        normalizedy = [] #[lineary[0]]

        for i in range(const_kmax):
            if linearx[i]/nbrgt < const_knmax:
                normalizedx.append(linearx[i]/nbrgt)
                normalizedy.append(lineary[i])

        flin = scipy.interpolate.interp1d(linearx, lineary, fill_value="extrapolate")
        fnorm = scipy.interpolate.interp1d(normalizedx, normalizedy, fill_value="extrapolate")
        funs.append(flin)
        normalizedFuns.append(fnorm)




    #     plt.figure(1)
    #     plt.scatter(normalizedx, normalizedy, marker=".", color=cmap(i*2))
    #     plt.figure(2)
    #     plt.scatter(linearx, lineary, marker=".", color=cmap(i*2+1))
    #
    # plt.figure(1)
    # fx2, fy2 = filter2(normalizedFuns,0,3,100)
    # plt.plot(fx2, fy2,  linewidth=2, color="red")
    # plt.figure(2)
    # fx2, fy2 = filter2(funs,1,100,100)
    # plt.plot(fx2, fy2,  linewidth=2, color="red")
    # plt.show()
    fx2, fy2 = filter2(normalizedFuns, const_knmin, const_knmax, const_knres)
    fx3, fy3 = filter2(funs,const_kmin,const_kmax,const_kres)
    return scipy.interpolate.interp1d(fx3, fy3, fill_value="extrapolate"), scipy.interpolate.interp1d(fx2, fy2, fill_value="extrapolate")

def computeStatFolderFloat(folder, imageIds):
    data = [];
    res = {}
    resDataset = {}
    res["dataset"] = resDataset
    with open(imageIds,"r") as idf:
        allids = idf.readlines()
    funsLin = []
    funsNorm = []

    countIm = len(allids)
    bestXLin = 0
    bestXNorm = 0
    bestYLin = 0
    bestYNorm = 0
    for imageid in allids:

        imageid = imageid.strip()
        #print(imageid)
        resim = {}
        resDataset[imageid] = resim
        filename = folder +  imageid +".json"

        if not os.path.isfile(filename):
            print("ERROR cannot find file: " + filename)
            exit(1)
        try:
            with open(filename) as data_file:
                data = json.load(data_file)
        except Exception as e:
            print("Error reading file " + filename +"\tMessage:" +str(e))
            exit(1)
        #print(data)
        resim["score_ogt"] = extract_best_match_GT(data)
        _,_,scoresbm,_ = zip(*resim["score_ogt"])

        resim["score_agt"] = extract_score_atGTRegions(data)
        _,_,scoresam,_ = zip(*resim["score_agt"])
        funLin, funNorm = compute_mean_curve(data)
        funsLin.append(funLin)
        funsNorm.append(funNorm)
        areaUnderCurveLin, bestScoreXLin, bestScoreYLin = computeScores(funLin,const_kmin,const_kmax,const_kres)
        areaUnderCurveNorm, bestScoreXNorm, bestScoreYNorm = computeScores(funNorm,const_knmin,const_knmax,const_knres)

        resim["mean_score_lin"] = [areaUnderCurveLin, bestScoreXLin, bestScoreYLin]
        resim["mean_score_norm"] = [areaUnderCurveNorm, bestScoreXNorm, bestScoreYNorm]
        resim["mean_score_ogt"] = sum(scoresbm)/len(scoresbm)
        resim["mean_score_agt"] = sum(scoresam)/len(scoresam)
        bestXLin += bestScoreXLin
        bestXNorm += bestScoreXNorm
        bestYLin += bestScoreYLin
        bestYNorm += bestScoreYNorm


        #resim["meanLinear"]=funLin
        #resim["meanNormalized"]=funsNorm


    bestXLin /= countIm
    bestXNorm /= countIm
    bestYLin /= countIm
    bestYNorm /= countIm

    mfunXNorm, mfunYNorm = filter2(funsNorm, const_knmin,const_knmax,const_knres)
    mfunXLin, mfunYLin = filter2(funsLin,const_kmin,const_kmax,const_kres)

    mfunLin = scipy.interpolate.interp1d(mfunXLin, mfunYLin, fill_value="extrapolate")
    mfunNorm = scipy.interpolate.interp1d(mfunXNorm, mfunYNorm, fill_value="extrapolate")

    areaUnderCurveLin, bestScoreXLin, bestScoreYLin = computeScores(mfunLin,const_kmin,const_kmax,const_kres)
    areaUnderCurveNorm, bestScoreXNorm, bestScoreYNorm = computeScores(mfunNorm,const_knmin,const_knmax,const_knres)

    xLin = numpy.linspace(const_kmin,const_kmax, num=const_kres, endpoint=True)
    yLin = mfunLin(xLin)
    res["curve_lin"] = list(zip(xLin,yLin))

    xNorm = numpy.linspace(const_knmin,const_knmax, num=const_knres, endpoint=True)
    yNorm = mfunNorm(xNorm)
    res["curve_norm"] = list(zip(xNorm,yNorm))

    res["ods_auc_lin"]= areaUnderCurveLin
    res["ods_best_lin"]= [bestScoreXLin, bestScoreYLin]
    res["ods_f(0.5)_lin"]= float(mfunLin(0.5))
    res["ods_f(1)_lin"]= float(mfunLin(1))
    res["ods_f(2)_lin"]= float(mfunLin(2))

    res["ods_auc_norm"]= areaUnderCurveNorm
    res["ods_best_norm"]= [bestScoreXNorm, bestScoreYNorm]
    res["ods_f(0.5)_norm"]= float(mfunNorm(0.5))
    res["ods_f(1)_norm"]= float(mfunNorm(1))
    res["ods_f(2)_norm"]= float(mfunNorm(2))

    res["ois_best_lin"]= [bestXLin, bestYLin]
    res["ois_best_norm"]= [bestXNorm, bestYNorm]

    return res
    # xx = []
    # yy = []
    # for e in resDataset.values():
    #     el = e["bestmatch"]
    #     for t in el:
    #         xx.append(t[0])
    #         yy.append(t[1])
    # plt.scatter(xx,yy)
    # plt.axis('equal')
    # plt.show()
    # # plt.figure(1)
    # # plt.plot(fx2, fy2, linewidth=2, color="red")
    # # plt.figure(2)
    # # plt.plot(fx3, fy3, linewidth=2, color="blue")
    # # plt.show()
    # print(json.dumps(res))


# def computeStatFolder(folder, imagesIds):
#
#     print(folder)
#     histo = {}
#     with open(imageIds,"r") as idf:
#         allids = idf.readlines()
#
#     header = ["#k\tscore_Mean"]
#
#     for imageid in allids:
#         imageid = imageid.strip()
#         filename = folder +  imageid +".txt"
#
#         if not os.path.isfile(filename):
#             print("ERROR cannot find file: " + filename)
#             exit(1)
#         with open(filename, newline='') as csvfile:
#             spamreader = csv.reader(csvfile, delimiter='\t')
#             spamreader.__next__() #skip header very dirty
#             for row in spamreader:
#                 k = int(row[0])
#                 l = histo.setdefault(k,[])
#                 l.append(float(row[1]))
#
#     result = [header]
#
#     sortedKeys = sorted(histo.keys())
#
#     for i in sortedKeys:
#         binvals = numpy.array(histo[i])
#         mv = scipy.mean(binvals)
#         result.append([i+1, mv])
#
#     return result
#
#
#computeStatFolderFloat(r"d:\travail\svnEsiee\ResTest\tmp\WSAreaPiotr.sh\bsds500\BCE\\", r"d:\travail\svnEsiee\ResTest\ids_test.txt" )
#computeStatFolderFloat(r"d:\travail\svnEsiee\ResTest\tmp\QFZFilteredPiotr.sh\bsds500\BCE\\", r"d:\travail\svnEsiee\ResTest\ids_test.txt" )

#computeStatFolderFloat(r"d:\travail\svnEsiee\ResTest\tmp\WSAreaPiotr.sh\bsds500\BCE_threshold\\", r"d:\travail\svnEsiee\ResTest\ids_test.txt" )

#computeStatFolderFloat(r"d:\travail\svnEsiee\ResTest\tmp\QFZFilteredPiotr.sh\bsds500\BCE_threshold\\", r"d:\travail\svnEsiee\ResTest\ids_test.txt" )

# exit()


if len(sys.argv) != 4:
    raise Exception("Incorrect usage: folder imagesIds outFile")
inFolder = sys.argv[1]
imageIds = sys.argv[2]
outFile = sys.argv[3]

result = computeStatFolderFloat(inFolder, imageIds)

with open(outFile, 'w') as outfile:
    json.dump(result, outfile)
